from typing import Union

try:
    from spatialpandas import GeoDataFrame
    from spatialpandas.dask import DaskGeoDataFrame

    _has_spatialpandas = True
except ImportError:
    _has_spatialpandas = False

try:
    import dask
    import dask.dataframe as dd
    from dask.distributed import Client, LocalCluster

    _has_dask = True
except ImportError:
    _has_dask = False

import geopandas as gpd

from cellseg_gsontools.spatial_context.ops import get_interface_zones, get_objs_sp

from ._base_backend import _SpatialBackend

if _has_dask:
    dask.config.set({"logging.distributed": "error"})

__all__ = ["_SpatialContextSP"]


class _SpatialContextSP(_SpatialBackend):
    def __init__(self) -> None:
        """Create a spatialpandas spatial context class."""
        if not _has_spatialpandas:
            raise ImportError(
                "spatialpandas is required to use the spatial context backend."
                "Install it with `pip install spatialpandas`."
            )
        self.backend_name = "spatialpandas"

    def roi(self, context_area: gpd.GeoDataFrame, ix: int) -> gpd.GeoDataFrame:
        """Get a roi area of index `ix`.

        Parameters
        ----------
        context_area : gpd.GeoDataFrame
            The context area gdf.
        ix : int
            The index of the roi area. I.e., the ith roi area.

        Returns
        -------
            gpd.GeoDataFrame:
                The ith roi area.
        """
        roi_area = context_area.loc[[ix]]

        return self.to_geopandas(roi_area)

    def roi_cells(
        self,
        roi_area: gpd.GeoDataFrame,
        cell_gdf: gpd.GeoDataFrame,
        predicate: str = "intersects",
        silence_warnings: bool = True,
        parallel: bool = False,
        num_processes: int = None,
        **kwargs,
    ) -> gpd.GeoDataFrame:
        """Get the cells within the roi area.

        Parameters
        ----------
        roi_area : gpd.GeoDataFrame
            The roi area.
        cell_gdf : GeoDataFrame
            The cell gdf.
        predicate : str, default="intersects"
            The predicate to use for the spatial join.
        silence_warnings : bool, default=True
            Flag, whether to silence all the geopandas related warnings.
        parallel : bool, default=False
            Whether to use parallel processing.
        num_processes : int, optional
            The number of processes/partitions to use for the DaskGeoDataFrame. If None,
            the number of partitions is set to the number of cores.

        Returns
        -------
            gpd.GeoDataFrame:
                The cells within the roi area.
        """
        # check to not compute roi_area if already computed
        if roi_area is None or roi_area.empty:
            return

        # distributing sjoin to many workers results in speedup only for large areas
        if roi_area.geometry.area.values[0] > 1e8 and parallel:
            with LocalCluster(
                n_workers=int(num_processes),
                processes=True,
                threads_per_worker=1,
            ) as cluster, Client(cluster) as client:  # noqa
                objs_within: GeoDataFrame = get_objs_sp(
                    GeoDataFrame(roi_area),
                    cell_gdf,
                    silence_warnings=silence_warnings,
                    predicate=predicate,
                )
        else:
            objs_within: GeoDataFrame = get_objs_sp(
                GeoDataFrame(roi_area),
                cell_gdf,
                silence_warnings=silence_warnings,
                predicate=predicate,
            )

        if objs_within is None or len(objs_within.index) == 0:
            return

        # rename spatial join columns
        objs_within = objs_within.rename(
            columns={
                "index_right": "spatial_context_id",
                "global_id_left": "global_id",
                "class_name_left": "class_name",
                "class_name_right": "spatial_context",
            },
        )

        # drop unnecessary columns and return
        objs_within = objs_within.drop(columns=["global_id_right"])

        return self.to_geopandas(objs_within)

    def interface(
        self,
        top_roi_area: gpd.GeoDataFrame,
        bottom_gdf: GeoDataFrame,
        buffer_dist: int,
    ) -> gpd.GeoDataFrame:
        """Get an interface area of index `ix`.

        Parameters
        ----------
            top_roi_area : gpd.GeoDataFrame
                The top roi area.
            bottom_gdf : gpd.GeoDataFrame
                The bottom gdf on top of which top_roi_area is buffered.
        """
        # Get the intersection of the roi and the area of type `label2`
        iface = get_interface_zones(
            buffer_area=top_roi_area,
            areas=bottom_gdf.to_geopandas().set_crs(
                epsg=4328, inplace=True, allow_override=True
            ),
            buffer_dist=buffer_dist,
        )

        # If there are many interfaces, dissolve them into one
        if len(iface) > 1:
            iface = iface.dissolve().set_geometry("geometry")

        return iface

    def to_geopandas(self, gdf) -> gpd.GeoDataFrame:
        """Convert a spatialpandas GeoDataFrame to a geopandas GeoDataFrame."""
        if isinstance(gdf, GeoDataFrame):
            gdf = (
                gdf.to_geopandas()
                .set_geometry("geometry")
                .set_crs(epsg=4328, inplace=True, allow_override=True)
            )
        return gdf

    def convert_area_gdf(self, gdf: gpd.GeoDataFrame) -> GeoDataFrame:
        """Convert area gdf to spatialpandas gdf."""
        if _has_spatialpandas:
            if isinstance(gdf, gpd.GeoDataFrame):
                gdf = GeoDataFrame(gdf)
            else:
                raise TypeError(
                    f"input gdf has to be a geopandas.GeoDataFrame. Got: {type(gdf)}"
                )
        else:
            raise ImportError(
                "spatialpandas is required to use the spatial context backend."
                "with backend == 'spatialpandas'. Install it with "
                "`pip install spatialpandas`."
            )

        return gdf

    def convert_cell_gdf(
        self, gdf: gpd.GeoDataFrame, parallel: bool, n_partitions: int = None
    ) -> Union[GeoDataFrame, DaskGeoDataFrame]:
        """Convert cell gdf to spatialpandas gdf.

        Parameters
        ----------
        gdf : gpd.GeoDataFrame
            The cell gdf to convert.
        parallel : bool
            Whether to convert to a DaskGeoDataFrame.
        n_partitions : int, optional
            The number of processes/partitions to use for the DaskGeoDataFrame. If None,
            the number of partitions is set to the number of cores.

        Returns
        -------
        Union[GeoDataFrame, DaskGeoDataFrame]:
            The converted cell gdf.
        """
        if _has_spatialpandas:
            if isinstance(gdf, gpd.GeoDataFrame):
                # spatialpandas can use only Point arrays in sjoin
                gdf["centroid"] = gdf.centroid
                gdf = gdf.set_geometry("centroid")
                gdf = GeoDataFrame(gdf)
            else:
                raise TypeError(
                    f"input gdf has to a geopandas.GeoDataFrame. Got: {type(gdf)}"
                )
        else:
            raise ImportError(
                "spatialpandas is required to use the spatial context backend."
                "with backend == 'spatialpandas'. Install it with "
                "`pip install spatialpandas`"
            )

        if parallel:
            if _has_dask and _has_spatialpandas:
                print(
                    f"partitioning cell_gdf into {n_partitions} partitions. This can "
                    "take a few minutes...",
                )
                gdf = dd.from_pandas(gdf, npartitions=n_partitions).persist()
                print("partitioning done.")

                # Precompute the partition-level spatial index
                gdf.partition_sindex
            else:
                raise ImportError(
                    "spatialpandas and dask is required to use the spatial context "
                    "backend. with backend == 'spatialpandas' and parallel==True. "
                    "Install with `pip install spatialpandas` and `pip install dask`."
                )
        return gdf



from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import geopandas as gpd
import pandas as pd
from shapely.geometry import MultiPolygon
from shapely.ops import unary_union
from tqdm import tqdm

from ..utils import set_uid
from ._base_merger import BaseGSONMerger
from .gson_tile import GSONTile
from .poly_utils import merge_overlaps, validate_and_simplify
from .save_utils import gdf_to_file

__all__ = ["CellMerger"]


class CellMerger(BaseGSONMerger):
    """Merge adjascent cell annotation files to one file.

    Note:
        Assumes:

        - Input files contain nuclei/cell instance segmentation annotations.
        - Files have start x and y coords embedded in filename e.g. `x-[coord]_y-[coord]`
        - Input Tiles are the same size.
        - Allowed input file-formats `.json`, `.geojson`, `.feather`, `.parquet`

    Parameters:
        in_dir (Union[Path, str]):
            Path to the directory containing the annotation files of tiles.
        tile_size (Tuple[int, int]):
            Height and width of the tiles in pixels.

    Attributes:
        border_annots (gpd.GeoDataFrame):
            A gdf of the merged border annotations. Available after merging.
        non_border_annots (gpd.GeoDataFrame):
            A gdf of the merged non-border annotations. Available after merging.
        annots (gpd.GeoDataFrame):
            A gdf of the resulting annotations. Available after merging.
    """

    def __init__(
        self, in_dir: Union[Path, str], tile_size: Tuple[int, int] = (1000, 1000)
    ) -> None:
        super().__init__(in_dir, tile_size)

        # Lookup to manage the relations between the main and adjacent tiles
        # main tile is the current tile and adj is the adjacent tile
        self.neighbor_relations = {
            "right": {"main": "right", "adj": "left"},
            "left": {"main": "left", "adj": "right"},
            "top": {"main": "top", "adj": "bottom"},
            "bottom": {"main": "bottom", "adj": "top"},
        }

        # Lookup for already visited neighbors
        self.visited = {
            f.name: {
                "left": None,
                "right": None,
                "top": None,
                "bottom": None,
                "non_border": None,
                "top_right": None,
                "top_left": None,
                "bottom_left": None,
                "bottom_right": None,
            }
            for f in self.files
        }

    def merge_dir(
        self,
        out_fn: Optional[Union[Path, str]] = None,
        format: Optional[str] = None,
        verbose: bool = True,
    ) -> None:
        """Merge all the instance segmentation files in the input directory.

        Parameters:
            out_fn (Union[Path, str]):
                Filename for the output file. If None, the merged gdf is saved to the
                class attribute `self.annots` only.
            format (str):
                The format of the output geojson file. One of: "feather", "parquet",
                "geojson", None. This is ignored if `out_fn` is None.
            verbose (bool):
                Whether to show a progress bar or not.

        Examples:
            Write geojson files to a standard '.geojson' file.
            >>> from cellseg_gsontools.merging import CellMerger
            >>> merger = CellMerger("/path/to/geojsons/", tile_size=(1000, 1000))
            >>> merger.merge_dir("/path/to/output.json", format="geojson")

            Write input geojson files to feather file.
            >>> from cellseg_gsontools.merging import CellMerger
            >>> merger = CellMerger("/path/to/geojsons/", tile_size=(1000, 1000))
            >>> merger.merge_dir("/path/to/output.feather", format="feather")

            Write input geojson files to parquet file.
            >>> from cellseg_gsontools.merging import CellMerger
            >>> merger = CellMerger("/path/to/geojsons/", tile_size=(1000, 1000))
            >>> merger.merge_dir("/path/to/output.parquet", format="parquet")
        """
        if out_fn is not None:
            out_fn = Path(out_fn)

        if format not in (".feather", ".parquet", ".geojson", None):
            raise ValueError(
                f"Invalid format. Got: {format}. Allowed: .feather, .parquet, .geojson"
            )

        # merge the tiles
        self.annots = self._merge(verbose=verbose)
        if verbose:
            msg = f"{format}-format" if out_fn is not None else "`self.annots`"
            print(f"Saving the merged geojson file: {out_fn} to {msg}")

        if out_fn is not None:
            # save the merged geojson
            gdf_to_file(self.annots, out_fn, format)

    def _get_non_border_polygons(self, gson: GSONTile) -> List[Dict[str, Any]]:
        """Get all the polygons that do not touch any edges of the tile.

        Parameters
        ----------
            gson : GSONTile
                GSONTile obj of a geojson tile.

        Returns
        -------
            Tuple[List[Polygon], List[str]]:
                A list of cell polygon objects and a list of the corresponding classes.
        """
        nb_annots = gson.non_border_annots

        classes = []
        new_polys = []
        if not nb_annots.empty:
            for poly, c in zip(nb_annots.geometry, nb_annots.class_name):
                poly = validate_and_simplify(poly, simplify=True, buffer=0.0)
                new_polys.append(poly)
                classes.append(c)

        return new_polys, classes

    def _merge_adj_ploygons(
        self, gson: GSONTile, gson_adj: GSONTile, adj_pos: str
    ) -> List[Dict[str, Any]]:
        """Merge adjascent polygons in two adjacsent geojsons.

        This concatenates the cells that are split at the image borders.

        Parameters
        ----------
            gson : GSONTile
                GSONTile obj of the geojson of the main tile.
            gson_adj : GSONTile
                GSONTile obj of the geojson of the adjascnet tile.
            adj_pos : str
                The postition of the adjascent tile relative to the main tile.
                One of: "left", "right", "bottom", "bottomleft", "bottomright", "top",
                "topleft", "topright"

        Returns
        -------
            Tuple[List[Polygon], List[str]]:
                A list of cell polygon objects and a list of the corresponding classes.
        """
        # Get the polygons that end/start at the image border
        if adj_pos == "right":
            border_annots_main = gson.right_border_annots
            border_annots_adj = gson_adj.left_border_annots
        elif adj_pos == "left":
            border_annots_main = gson.left_border_annots
            border_annots_adj = gson_adj.right_border_annots
        elif adj_pos == "bottom":
            border_annots_main = gson.bottom_border_annots
            border_annots_adj = gson_adj.top_border_annots
        elif adj_pos == "top":
            border_annots_main = gson.top_border_annots
            border_annots_adj = gson_adj.bottom_border_annots

        # combine polygons that intersect/touch between two image tiles
        # (cells that are split in two between two image tiles)
        new_classes = []
        new_polys = []
        if not border_annots_main.empty and not border_annots_adj.empty:
            for main_poly, c1 in zip(
                border_annots_main.geometry, border_annots_main.class_name
            ):
                main_poly = validate_and_simplify(main_poly, buffer=0.0)
                for adj_poly, c2 in zip(
                    border_annots_adj.geometry, border_annots_adj.class_name
                ):
                    adj_poly = validate_and_simplify(adj_poly, buffer=0.0)

                    # combine the polygons if they intersect
                    if main_poly.intersects(adj_poly):
                        new_poly = unary_union([main_poly, adj_poly])

                        # do some simplifying
                        new_poly = validate_and_simplify(
                            new_poly, simplify=True, buffer=0.0
                        )

                        if isinstance(new_poly, MultiPolygon):
                            print("has multipoly")

                        # take the class of the larger object
                        if adj_poly.area >= main_poly.area:
                            new_class = c2
                        else:
                            new_class = c1

                        new_polys.append(new_poly)
                        new_classes.append(new_class)

        return new_polys, new_classes

    def _merge(self, verbose: bool = True) -> gpd.GeoDataFrame:
        """Merge all annotations in the files to one.

        Handles the split cells at the image borders.

        NOTE: This does not handle corners
        """
        non_border_annots = []
        non_border_classes = []
        border_annots = []
        border_classes = []
        pbar = tqdm(self.files) if verbose else self.files
        for f in pbar:
            if verbose:
                pbar.set_description(f"Processing file: {f.name}")

            # get adjascent tiles
            adj = self._get_adjascent_tiles(f.name)

            # Init GSONTile obj
            gson = GSONTile(f, tile_size=self.tile_size)
            if gson.gdf is not None and not gson.gdf.empty:
                # add the non border polygons
                if self.visited[f.name]["non_border"] is None:
                    non_border_polygons, non_border_cls = self._get_non_border_polygons(
                        gson
                    )
                    non_border_annots.extend(non_border_polygons)
                    non_border_classes.extend(non_border_cls)
                    self.visited[f.name]["non_border"] = f.name

                # loop the adjascent tiles and add the border polygons
                for pos, f_adj in adj.items():
                    if f_adj is not None:
                        if self.visited[f.name][pos] is None:
                            gson_adj = GSONTile(f_adj, tile_size=self.tile_size)

                            if gson_adj.gdf is not None and not gson_adj.gdf.empty:
                                border_polygons, border_cls = self._merge_adj_ploygons(
                                    gson, gson_adj, pos
                                )
                                border_annots.extend(border_polygons)
                                border_classes.extend(border_cls)

                                # update lookup
                                main_pos = self.neighbor_relations[pos]["main"]
                                adj_pos = self.neighbor_relations[pos]["adj"]
                                self.visited[f_adj.name][adj_pos] = f.name
                                self.visited[f.name][main_pos] = f_adj.name

        # save the annotations to class attributes
        self.non_border_annots = gpd.GeoDataFrame(
            {"geometry": non_border_annots, "class_name": non_border_classes}
        )
        border_annots = gpd.GeoDataFrame(
            {"geometry": border_annots, "class_name": border_classes}
        )
        self.border_annots = merge_overlaps(border_annots)  # merge overlapping objects
        annots = set_uid(
            pd.concat([self.non_border_annots, self.border_annots]),
            0,
            id_col="id",
            drop=False,
        )
        return annots[annots.class_name != "background"]



from typing import Any, Dict, List, Tuple

import geopandas as gpd
from matplotlib import pyplot as plt
from matplotlib.cm import get_cmap

__all__ = ["plot_all", "plot_gdf"]


def replace_legend_items(legend, mapping):
    for txt in legend.texts:
        for k, v in mapping.items():
            if txt.get_text() == str(k):
                txt.set_text(v)


def plot_gdf(
    gdf: gpd.GeoDataFrame,
    col: str,
    ax: plt.Axes = None,
    cmap: str = None,
    bin_legends: List[str] = None,
    n_bins: int = None,
    show_legend: bool = True,
    loc: str = "upper right",
    figsize: tuple = (10, 10),
    **kwargs,
) -> plt.Axes:
    """Plot one gdf wrapper.

    Parameters:
        gdf (gpd.GeoDataFrame):
            The gdf to plot.
        col (str):
            The column to highlight.
        ax (plt.Axes):
            The axes to plot on, by default None.
        cmap (str):
            The colormap to use, by default None.
        bin_legends (List[str]):
            The bins to use, by default None.
        n_bins (int):
            The number of bins to use, by default None.
        show_legend (bool):
            Whether to show the legend, by default True.
        loc (str):
            The location of the legend, by default "upper right".
        figsize (tuple):
            The size of the figure, by default (10, 10).
        **kwargs (Dict[str, Any])):
            Extra keyword arguments passed to the `plot` method of the GeoDataFrame.

    Returns:
        plt.Axes:
            The axes used for plotting.
    """
    if cmap is not None:
        cmap = get_cmap(cmap, n_bins)

    ax = gdf.plot(
        ax=ax,
        column=col,
        cmap=cmap,
        categorical=True,
        legend=show_legend,
        legend_kwds={"loc": loc},
        figsize=figsize,
        **kwargs,
    )
    if show_legend:
        leg = ax.legend_
        ax.add_artist(leg)

    if cmap is not None and show_legend and bin_legends is not None:
        mapping = dict([(i, s) for i, s in enumerate(bin_legends)])
        replace_legend_items(ax.get_legend(), mapping)

    return ax


def plot_all(
    area_gdf: gpd.GeoDataFrame,
    cell_gdf: gpd.GeoDataFrame,
    context_gdf: gpd.GeoDataFrame = None,
    network_gdf: gpd.GeoDataFrame = None,
    grid_gdf: gpd.GeoDataFrame = None,
    show_legends: bool = True,
    color: str = None,
    figsize: Tuple[int, int] = (12, 12),
    grid_cmap: str = None,
    grid_col: str = None,
    grid_n_bins: int = None,
    edge_kws: Dict[str, Any] = None,
    **kwargs,
) -> plt.Axes:
    """Plot the areas, cells, and spatial context areas together.

    Parameters:
        area_gdf (gpd.GeoDataFrame):
            GeoDataFrame containing the areas of interest.
        cell_gdf (gpd.GeoDataFrame):
            GeoDataFrame containing the cells.
        context_gdf (gpd.GeoDataFrame):
            GeoDataFrame containing the context, by default None.
        network_gdf (gpd.GeoDataFrame):
            GeoDataFrame containing the network, by default None.
        grid_gdf (gpd.GeoDataFrame):
            GeoDataFrame containing the grid, by default None.
        show_legends (bool):
            Flag, whether to include legends for each in the plot.
        color (str):
            A color for the interfaces or rois, Ignored if `show_legends=True`.
        figsize (Tuple[int, int]):
            Size of the figure.
        grid_cmap (str):
            The colormap to use for the grid. If `grid_gdf` is None, this is ignored.
        grid_col (str):
            The column to use for the grid. If `grid_gdf` is None, this is ignored.
        grid_n_bins (int):
            The number of bins to use for the grid. If `grid_gdf` is None, this is ignored.
        **kwargs (Dict[str, Any])):
            Extra keyword arguments passed to the `plot` method of the geodataframes.

    Returns:
        AxesSubplot:
            The axes used for plotting.
    """
    _, ax = plt.subplots(figsize=figsize)

    ax = area_gdf.plot(
        ax=ax,
        column="class_name",
        categorical=True,
        legend=show_legends,
        alpha=0.1,
        legend_kwds={
            "loc": "upper center",
        },
        aspect=None,
        **kwargs,
    )
    if show_legends:
        leg1 = ax.legend_
        ax.add_artist(leg1)

    ax = cell_gdf.plot(
        ax=ax,
        column="class_name",
        categorical=True,
        legend=show_legends,
        legend_kwds={
            "loc": "upper right",
        },
        aspect=None,
        **kwargs,
    )
    if show_legends:
        leg2 = ax.legend_
        ax.add_artist(leg2)

    if context_gdf is not None:
        ax = context_gdf.plot(
            ax=ax,
            color=color,
            column="label",
            alpha=0.7,
            legend=False,
            categorical=True,
            aspect=None,
            **kwargs,
        )

    if network_gdf is not None:
        edge_kws = edge_kws or {}
        ax = network_gdf.plot(
            ax=ax,
            column="class_name",
            categorical=True,
            legend=show_legends,
            legend_kwds={"loc": "upper left"},
            cmap="Paired",
            **edge_kws,
        )
        if show_legends:
            leg3 = ax.legend_
            ax.add_artist(leg3)

    if grid_gdf is not None:
        cmap = None
        if color is not None or grid_cmap is not None:
            if grid_n_bins is not None:
                cmap = get_cmap(grid_cmap, grid_n_bins)
            else:
                cmap = get_cmap(grid_cmap)

        grid_gdf.geometry = grid_gdf.geometry.boundary
        ax = grid_gdf.plot(
            ax=ax,
            color=color,
            alpha=0.7,
            cmap=cmap,
            column=grid_col,
            categorical=True,
            legend=show_legends,
            legend_kwds={"loc": "lower left"},
        )
        if show_legends:
            leg4 = ax.legend_
            ax.add_artist(leg4)

    return ax



from ._base_backend import _SpatialBackend
from ._dgp_backend import _SpatialContextDGP
from ._gp_backend import _SpatialContextGP
from ._sp_backend import _SpatialContextSP

__all__ = [
    "_SpatialBackend",
    "_SpatialContextGP",
    "_SpatialContextSP",
    "_SpatialContextDGP",
]
